var gTreeView;
var gFilterBox;
var gTree;
var gDeck;
var gFetchInProgress;
var gRcCmd;

var NS_HRC = "@merrillpress.com/2008/hrc";
var NS_XHTML = "http://www.w3.org/1999/xhtml";
var NS_XUL = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";

var COV_MEDIOCRE = 50; /* >= */
var COV_GOOD     = 75; /* >= */
var COV_GREAT    = 95; /* >= */

var PAGE_RECORDING  = 0;
var PAGE_PROCESSING = 1;
var PAGE_RESULTS    = 2;

function trans_results_to_recording() {
  assert(function() gDeck.selectedIndex == PAGE_RESULTS);
  gDeck.selectedIndex = PAGE_RECORDING;
  force_redraw();

  gHrc.clearData();
  gHrc.start();
  gRcCmd.setAttribute('checked', 'true');
}

function trans_recording_to_processing() {
  assert(function() gDeck.selectedIndex == PAGE_RECORDING);
  assert(function() !gFetchInProgress);

  gDeck.selectedIndex = PAGE_PROCESSING;
  force_redraw();
  
  gHrc.stop();
  gRcCmd.setAttribute('checked', 'false');
  gRcCmd.setAttribute('disabled', 'true');

  var raw_results = gHrc.getData({});
  var sources_to_get = raw_results.map(function(result) result.uri);

  function on_fetch_done(raw_sources) {
    trans_processing_to_results(raw_results, raw_sources);
  }

  gFetchInProgress = new ASyncURLFetcher(
    sources_to_get, on_fetch_done);
}

function trans_processing_to_results(raw_results, raw_sources) {
  assert(function() gDeck.selectedIndex == PAGE_PROCESSING);
  assert(function() gFetchInProgress);
  
  gFetchInProgress = null;
  var sources = {};
  for(var uri in raw_sources) {
    var source = raw_sources[uri];
    if(source !== null) {
      source = source.split(/\r?\n/);
    }

    sources[uri] = source;
  }

  var results = [];
  
  for(var i = 0; i < raw_results.length; ++i) {
    var raw_result = raw_results[i];
    var uri = raw_result.uri;

    var e = raw_result.executedLines;
    var p = raw_result.potentialLines;

    e.matchSize(p);

    var u = p.clone();
    var not_e = e.clone(); not_e.flip();
    u.and(not_e);

    var nr_p = p.numberBitsSet;
    var nr_u = u.numberBitsSet;
    var percent = 100 * (nr_p - nr_u) / nr_p;

    results.push({ uri:             uri,
                   lines:           p,
                   unExecLines:     u,
                   nrUnExecLines:   nr_u,
                   nrLines:         nr_p,
                   nrExecLines:     nr_p - nr_u,
                   coveragePercent: percent,
                   source:          sources[uri] || null
                 });
  }
  
  gRcCmd.setAttribute('disabled', 'false');
  gTreeView.update(results);
  gDeck.selectedIndex = PAGE_RESULTS;
}

function on_filter_input() {
  var regexp_s = gFilterBox.value;

  if(regexp_s === '') {
    regexp_s = '.*';
  }

  try {
    var regexp = RegExp(regexp_s);
  } catch(ex) {
    gFilterBox.setAttribute('class', 'regexpError');
    return;
  }

  gFilterBox.setAttribute('class', '');
  gTreeView.setFilter(regexp);
}

function view_source(url, lineno_start, lineno_end) {
  openDialog("chrome://global/content/viewSource.xul",
             "_blank",
             "all,dialog=no",
             url, null, null, lineno_start, false);
}

function sort_data(data, sortby) {
  var key = sortby.substring(4);
  if(key === 'coverage') {
    key = 'coveragePercent';
  }

  function sorter(a, b) {
    var val1 = a[key];
    var val2 = b[key];

    if(val1 < val2) {
      return -1;
    }

    if(val2 < val1) {
      return 1;
    }

    return 0;
  }

  data.sort(sorter);
}

function copy_list(aList) {
  return aList.map(function(x) x);
}

/**
  * Calls func for each selected row in a treeview. The argument is
  * the row number.
  */
function for_selected_rows(treeview, func) {
  var selection = gTree.view.selection;
  var num_ranges = selection.getRangeCount();
  var min = {};
  var max = {};

  for(var rangeno = 0; rangeno < num_ranges; ++rangeno) {
    selection.getRangeAt(rangeno, min, max);
    for(var rowno = min.value; rowno <= max.value; ++rowno) {
      func(rowno);
    }
  }
}

var last_selected_rows = [];

function on_select() {
  dump('on_select\n');
  var details = document.getElementById('details');
  var selection = window.getSelection();
  clear_children(details);
  for_selected_rows(gTree, function(rowno) {
    var detail = add_detail_xml(
      gTreeView._data[rowno], details);
  });
}

function ranges_from_bitset(bs) {
  var ranges = [];
  var open = null;
  var size = bs.size;

  for(var i = 0; i < size; ++i) {
    if(bs.testBit(i)) {
      if(open === null) {
        open = i;
      }
    } else {
      if(open !== null) {
        ranges.push([open, i - 1]);
        open = null;
      }
    }
  }

  if(open !== null) {
    ranges.push([open, size - 1]);
  }

  return ranges;
}

function on_source_range_clicked(e) {
  e.preventDefault();

  var row = e.target.row;
  var nc_range = e.target.nc_range;
  view_source(row.uri, nc_range[0], nc_range[1]);
}

function on_sourcett_showing(e) {
  var row = document.tooltipNode.row;
  var nc_range = document.tooltipNode.nc_range;

  var f = this.firstChild;
  clear_children(f);

  var rs = row.source;

  if(rs && rs.length) {
    var first = Math.max(1, nc_range[0] - 3);
    var last = Math.min(nc_range[1] + 3, rs.length);

    var max_digits = (last + '').length + 1;

    for(var i = first; i <= last; ++i) {
      var lineno_string = i + ':';
      while(lineno_string.length < max_digits) {
        lineno_string = ' ' + lineno_string;
      }

      var p = f;

      if(i >= nc_range[0] && i <= nc_range[1]) {
        p = document.createElementNS(NS_XHTML, 'span');
        p.className = 'srcinregion';
        f.appendChild(p);
      }

      p.appendChild(document.createTextNode(lineno_string + rs[i-1] + '\n'));
    }
  } else {
    f.appendChild(document.createTextNode('[source unavailable]'));
  }
}

function add_detail_xml(row, parent) {
  var detail = document.createElementNS(NS_HRC, 'detail');

  var uri = document.createElementNS(NS_HRC, 'uri');
  uri.appendChild(document.createTextNode(row.uri));
  detail.appendChild(uri);

  function add(key_name) {
    var kvpair = document.createElementNS(NS_HRC, 'kvpair');
    var key = document.createElementNS(NS_HRC, 'key');
    key.appendChild(document.createTextNode(key_name));
    var value = document.createElementNS(NS_HRC, 'value');
    kvpair.appendChild(key);
    kvpair.appendChild(value);
    detail.appendChild(kvpair);
    return value;
  }
  
  if(row.nrUnExecLines) {
    var nc = add('Lines not run: ');
    var nc_ranges = ranges_from_bitset(row.unExecLines);
    
    for(var i = 0;  i < nc_ranges.length; ++i) {
      if(i !== 0) {
        nc.appendChild(document.createTextNode(', '));
      }

      var nc_range = nc_ranges[i];
      if(nc_range[0] === nc_range[1]) {
        var nc_range_text = nc_range[0];
      } else {
        var nc_range_text = nc_range[0] + '\u2013' + nc_range[1];
      }

      link = document.createElementNS(NS_XHTML, 'a');
      link.href = '#';
      link.row = row;
      link.nc_range = nc_range;
      link.addEventListener('click', on_source_range_clicked, false);
      link.setAttribute('tooltip', 'sourcett');
      link.appendChild(document.createTextNode(nc_range_text));
      nc.appendChild(link);
    }
  }
  
  parent.appendChild(detail);
  return detail;
}

function HRCovTreeView() {
  this._rawData = this._data = [];
  this.rowCount = 0;
  this._filter = /.*/;
}

HRCovTreeView.prototype = {
  constructor: HRCovTreeView,

  _getTree: function() {
    return this._treebox.treeBody.parentNode;
  },

  _doRefresh: function() {
    var tree = this._getTree();
    var results = [];
    
    for(var i = 0; i < this._rawData.length; ++i) {
      var item = this._rawData[i];
      if(this._filter.test(item.uri)) {
        results.push(item);
      }
    }

    var [sort_resource, sort_direction] =
      update_tree_sorting(tree);
    
    sort_data(results, sort_resource);
    
    if(sort_direction !== 'ascending') {
      results.reverse();
    }

    var selection = tree.view.selection;
    selection.selectEventsSuppressed = true;
    selection.clearSelection();
    this._data = results;
    var old_row_count = this.rowCount;
    this.rowCount = this._data.length;
    this._treebox.rowCountChanged(0, this.rowCount - old_row_count);
    selection.selectEventsSuppressed = false;
  },

  setFilter: function(regexp) {
    this._filter = regexp;
    this._doRefresh();
  },

  /* Update with new data, preserving selection */
  update: function(data) {
    this._rawData = data;
    this._doRefresh();
  },

  getProgressMode: function(row, column) {
    return Ci.nsITreeView.PROGRESS_NORMAL;
  },

  getCellText: function(row, column) {
    var id = column.id;
    if(id === 'col_uri') {
      return this._data[row].uri;
    } else if(id === 'col_coveragePercent') {
      return this._data[row].coveragePercent.toPrecision(3)
        + '%';
    } else if(id === 'col_nrExecLines') {
      return this._data[row].nrExecLines;
    } else if(id === 'col_nrUnExecLines') {
      return this._data[row].nrUnExecLines;
    } else if(id === 'col_nrLines') {
      return this._data[row].nrLines;
    }
  },

  getCellValue: function(row, column) {
    return Math.round(this._data[row].coveragePercent);
  },

  cycleHeader: function(column) {
    var tree = this._treebox.treeBody.parentNode;
    
    if(tree.getAttribute('sortResource') === column.id) {
      if(tree.getAttribute('sortDirection') === 'ascending') {
        tree.setAttribute('sortDirection', 'descending');
      } else {
        tree.setAttribute('sortDirection', 'ascending');
      }
    } else {
      tree.setAttribute('sortResource', column.id);
      tree.setAttribute('sortDirection', 'ascending');
    }

    this._doRefresh();
  },

  setTree: function(treebox) {
    this._treebox = treebox;
  },

  isContainer: function(row) {
    return false;
  },

  isSeparator: function(row) {
    return false;
  },

  isSorted: function() {
    return true;
  },

  getLevel: function() {
    return 0;
  },

  getImageSrc: function(row, column) {
    return null;
  },

  getRowProperties: function(row, props) {},

  
  getCellProperties: function(row, column, props) {
    if(column.id === 'col_coverage') {
      var aserv=Cc["@mozilla.org/atom-service;1"].
        getService(Ci.nsIAtomService);

      var percent = this._data[row].coveragePercent;

      if(percent >= COV_GREAT) {
        props.AppendElement(aserv.getAtom('greatCoverage'));
      } else if(percent >= COV_GOOD) {
        props.AppendElement(aserv.getAtom('goodCoverage'));
      } else if(percent >= COV_MEDIOCRE) {
        props.AppendElement(aserv.getAtom('mediocreCoverage'));
      } else if(percent) {
        props.AppendElement(aserv.getAtom('badCoverage'));
      }
    }
  },
  
  getColumnProperties: function(colid, col, props) {}
};

function results_onload() {
  gDeck = document.getElementById('deck');
  gTree = document.getElementById('results-tree');
  gRcCmd = document.getElementById('hrc_toggle_recording');
  gFilterBox = document.getElementById('filter');

  gTree.view = gTreeView = new HRCovTreeView();
  on_filter_input();
  
  document.getElementById('sourcett')
    .addEventListener('popupshowing', on_sourcett_showing, false);
  gTree.addEventListener('select', on_select, false);
  gFilterBox.addEventListener('input', on_filter_input, false);
}

function results_onunload() {
  if(gFetchInProgress) {
    gFetchInProgress.cancel();
    gFetchInProgress = null;
  }
}

function hrc_toggle_recording() {
  if(gDeck.selectedIndex == PAGE_RESULTS) {
    trans_results_to_recording();
  } else if(gDeck.selectedIndex == PAGE_RECORDING) {
    trans_recording_to_processing();
  }
}
